import gnnenrich
import pytest
import random

def test_RandomListsGenerator():

    # Empty list
    gen = gnnenrich.RandomListsGenerator([], 0)
    assert gen.size == 0
    assert next(gen) == []
    with pytest.raises(ValueError):
        gen = gnnenrich.RandomListsGenerator([], 1)

    # One item, size 0
    gen = gnnenrich.RandomListsGenerator([1], 0)
    assert gen.size == 0
    assert next(gen) == []
    with pytest.raises(ValueError):
        gen = gnnenrich.RandomListsGenerator([1], 2)

    # One item, size 1
    gen = gnnenrich.RandomListsGenerator([1], 1)
    assert gen.size == 1
    assert gen.length == -1
    assert gen.offset == 0
    assert next(gen) == [1]
    assert next(gen) == [1]
    assert gen.count == 2

    # Two items, size 1
    random.seed(42)
    gen = gnnenrich.RandomListsGenerator([1, 2], 1)
    assert gen.size == 1
    assert gen.length == -1
    assert next(gen) == [1]
    assert next(gen) == [1]
    assert next(gen) == [2]
    assert gen.count == 3
    
def test_keep_generated_lists():

    random.seed(42)
    gen = gnnenrich.RandomListsGenerator([1, 2], 1)
    assert next(gen) == [1]
    assert gen.get_lists() == []

    random.seed(42)
    gen = gnnenrich.RandomListsGenerator([1, 2], 1, keep=True)
    assert next(gen) == [1]
    assert next(gen) == [1]
    assert next(gen) == [2]
    assert gen.get_lists() == [[1], [1], [2]]
    
def test_offset():
    random.seed(42)
    gen = gnnenrich.RandomListsGenerator([1, 2], 1, offset=1)
    assert gen.offset == 1
    assert next(gen) == [1]
    assert next(gen) == [2]

def test_length():
    random.seed(42)
    gen = gnnenrich.RandomListsGenerator([1, 2], 1, length=2)
    assert gen.length == 2
    assert next(gen) == [1]
    assert next(gen) == [1]
    assert gen.count == 2
    with pytest.raises(StopIteration):
        assert next(gen) == [2]
    assert gen.count == 2
    
def test_generate_random_lists():
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a'], 1, 1) == [['a']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b'], 1, 1) == \
        [['a']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b'], 2, 1) == \
        [['a'], ['a']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b'], 1, 2) == \
        [['a', 'b']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b', 'c'], 1, 3) == \
        [['c', 'a', 'b']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b', 'c'], 1, 2) == \
        [['c', 'a']]
    random.seed(42)
    assert gnnenrich.generate_random_lists(['a', 'b', 'c'], 2, 2) == \
        [['c', 'a'], ['a', 'b']]



import collections
import gnnenrich
import numpy
import pytest

def test_get_items_embeddings():

    emb = collections.OrderedDict()
    emb['A'] = [1.0, 2.1]
    emb['B'] = [0.5, 3.4]

    scoring = gnnenrich.Scoring(emb)
    assert numpy.array_equal(scoring.get_items_embeddings(set(['A'])),
                             numpy.array([[1.0, 2.1]], dtype=numpy.float32))
    
def test_wrong_embeddings():

    emb = collections.OrderedDict()
    emb['A'] = [1.0, 2.1]
    emb['B'] = [0.5, 3.4, 7.8] # Embedding with different size

    with pytest.raises(ValueError):
        scoring = gnnenrich.Scoring(emb)



"""Random lists generation."""
# pylint: disable=unknown-option-value
from __future__ import annotations # For list[str] type hint in 3.7
import random
import typing

class RandomListsGenerator:
    """Generator of random lists.

    Generates random lists of string items (usually proteins or genes).

    Args:
        items: An iterable containing the string items.
        size: The size of each generated list.
        keep: When sets to True, stores the generated lists, that can be later
              retrieved with the get_lists() method.
        offset: If > 0, start by throwing away this number of random lists. This
                is useful for testing purposes, when using a single seed for
                random numbers and running in parallel, making sure each process
                or thread generates a unique set of random lists.
        length: If negative, never stops generating random lists. If positive,
                stops after having generated this number of lists.
    """

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, items: typing.Iterable[str], size: int,
                 keep: bool=False, offset: int=0, length: int=-1) -> None:

        self._items = list(items)
        self._size = size
        self._keep = keep
        self._offset = offset
        self._length = length
        self._count = 0
        if self._keep:
            self._lists: typing.List[typing.List[str]] = []

        if self._size > len(self._items):
            raise ValueError(f"The number of elements to sample ({self._size})"
                + " is greater than the number of available elements"
                + " ({len(self._items)}).")

        # Skip first lists
        for _ in range(self._offset):
            random.sample(self._items, self._size)

    @property
    def length(self) -> int:
        """The number of random lists configured."""
        return self._length

    @property
    def count(self) -> int:
        """The number of random lists generated up to now."""
        return self._count

    @property
    def offset(self) -> int:
        """The configured offset."""
        return self._offset

    @property
    def size(self) -> int:
        """The configured size."""
        return self._size

    def __iter__(self) -> 'RandomListsGenerator':
        """Gets the iterator on the random lists.

        Returns:
            Itself as an iterator.
        """
        return self

    def __next__(self) -> list[str]:
        """Generates a random list.
        
        Returns:
            A random list of <size> items.
        """

        # Done
        if self._length >= 0 and self._count >= self._length:
            raise StopIteration

        # Generate a new list
        lst = random.sample(self._items, self._size)
        self._count += 1

        # Keep the new list in memory
        if self._keep:
            self._lists.append(lst)

        return lst

    def get_lists(self) -> typing.List[typing.List[str]]:
        """Gets the generated random lists.
        
        If <keep> was set to True, returns the generated random lists, otherwise
        returns None.

        Returns:
            A list of all generated random lists up to now.
        """
        if self._keep:
            return self._lists
        return []

def generate_random_lists(items: list[str], n: int, m: int) -> \
        list[list[str]]:
    """Generates a series of random lists.

    Takes a list of string items, and randomly build n lists of m items.

    Args:
        items: an iterable containing the items.
        n: number of lists.
        m: the size of each generated list.
        
    Returns:
        A list of n lists of m items randomly chosen between the provided items.
    """
    random_lists = []
    gen = RandomListsGenerator(items, size=m, length=n)

    # Generate n lists
    for lst in gen:
        random_lists.append(lst)

    return random_lists



import collections
import gnnenrich
import numpy
import pytest

def test_comp_corr():

    emb = collections.OrderedDict()

    scoring = gnnenrich.Scoring(emb)
    assert scoring.corr_threshold == 0.95
    a = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[1, 1, 2, 3], [1, 1, -4, -5]])
    assert scoring.comp_corr(a, b) == 0.0

    scoring = gnnenrich.Scoring(emb, corr_threshold=0.9)
    assert scoring.corr_threshold == 0.9
    a = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[1, 1, 2, 3], [1, 1, -4, -5]])
    assert scoring.comp_corr(a, b) == 1.0

    scoring = gnnenrich.Scoring(emb, corr_threshold=0.9,
                                corr_threshold_repl_value=0.0)
    assert scoring.corr_threshold == 0.9
    a = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[1, 1, 2, 3], [1, 1, -4, -5]])
    assert scoring.comp_corr(a, b) == 0.9356646608879388


    scoring = gnnenrich.Scoring(emb, corr_threshold=0.9,
                                corr_threshold_repl_value=0.0,overlap_score=10)
    assert scoring.corr_threshold == 0.9
    assert scoring.overlap_score == 10
    a = numpy.array([[0, 1, 2, 3],[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[1, 1, 2, 3],[0, 1, 2, 3], [1, 1, -4, -5]])
    assert scoring.comp_corr(a, b) == 3.9571097739252925333333333


    scoring = gnnenrich.Scoring(emb, corr_threshold=0.95,
                                overlap_score=10)
    assert scoring.corr_threshold == 0.95
    assert scoring.overlap_score == 10
    a = numpy.array([[0, 1, 2, 3],[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[1, 1, 2, 3],[0, 1, 2, 3], [1, 1, -4, -5]])
    assert scoring.comp_corr(a, b) == 3.3333333333333333333333333
    
def test_overlap_score():

    emb = collections.OrderedDict()

    scoring = gnnenrich.Scoring(emb)
    assert scoring.overlap_score == 3.0

    scoring = gnnenrich.Scoring(emb)
    a = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    assert scoring.comp_corr(a, b) == 3.0

    overlap = 4.0
    scoring = gnnenrich.Scoring(emb, overlap_score=overlap)
    a = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    b = numpy.array([[0, 1, 2, 3], [1, 0, -1, -2]])
    assert scoring.comp_corr(a, b) == overlap



import collections
import gnnenrich
import numpy
import pytest

def test_initializer() -> None:

    emb = collections.OrderedDict()

    for v in (0.0, 0.1, 0.9999999999999999, -1):
        with pytest.raises(ValueError):
            gnnenrich.Scoring(emb, overlap_score = v)

    _ = gnnenrich.Scoring(emb, overlap_score = 1.0)

def test_set_pathways() -> None:

    # Create Scoring object
    emb = collections.OrderedDict()
    scoring = gnnenrich.Scoring(emb)

    # Try to get pathway keys
    with pytest.raises(RuntimeError):
        _ = scoring.get_pathway_keys()

    # Set pathways
    pw: gnnenrich.PathwaysDict = {"pw1": ["a", "b"]}
    scoring.set_pathways(pw)

    # Get pathway keys
    assert scoring.get_pathway_keys() == ["pw1"]

def test_get_pathway_embedding() -> None:

    # Create Scoring object
    emb: gnnenrich.Embeddings = collections.OrderedDict([
            ("a", [0.1, 0.4]),
            ("b", [0.3, 1.0]),
            ])
    scoring = gnnenrich.Scoring(emb)

    # Try to get embeddings
    with pytest.raises(RuntimeError):
        _ = scoring.get_pathway_embeddings()

    # Set pathways
    pw: gnnenrich.PathwaysDict = {"pw1": ["a", "b"]}
    scoring.set_pathways(pw)

    # Get embeddings
    emb = scoring.get_pathway_embeddings() 
    assert isinstance(emb, dict)
    assert len(emb) == 1
    assert "pw1" in emb

    # Get pathway embedding by key
    assert (numpy.array(scoring.get_pathway_embedding("pw1")) ==
            emb["pw1"]).all()
    with pytest.raises(KeyError):
        _ = scoring.get_pathway_embedding("pw2")

def test_compute_pathway_score() -> None:

    # Create Scoring object
    emb: gnnenrich.Embeddings = collections.OrderedDict([
            ("a", [0.1, 0.4]),
            ("b", [0.3, 1.0]),
            ])
    scoring = gnnenrich.Scoring(emb)

    # Set pathways
    pw: gnnenrich.PathwaysDict = {"pw1": ["a", "b"]}
    scoring.set_pathways(pw)

    # Test compute_pathway_score()
    e: EmbArr = numpy.array([[1.0, 2.0], [0.4, 0.7]])
    score = scoring.compute_pathway_score("pw1", e)
    assert isinstance(score, float)
    assert score == 3.0
    with pytest.raises(KeyError):
        _ = scoring.compute_pathway_score("pw2", e)

def test_compute_pathway_scores() -> None:

    # Create Scoring object
    emb: gnnenrich.Embeddings = collections.OrderedDict([
            ("a", [0.1, 0.4]),
            ("b", [0.3, 1.0]),
            ])
    scoring = gnnenrich.Scoring(emb)

    # Set pathways
    pw: gnnenrich.PathwaysDict = {"pw1": ["a", "b"]}
    scoring.set_pathways(pw)

    # Test compute_pathway_scores()
    e: EmbArr = numpy.array([[1.0, 2.0], [0.4, 0.7]])
    scoring.compute_pathway_scores(e)
    assert {"pw1": 3.0} == scoring.get_computed_pathway_scores()



"""GNN Enrich scoring module."""
# pylint: disable=unknown-option-value
import collections
import typing
import logging
import numpy
import numpy.typing
import scipy.spatial.distance # type: ignore[import-untyped]
import tqdm # type: ignore[import-untyped]

# Get logger
logger = logging.getLogger('gnnenrich')

# Types
PathwaysDict = dict[str, list[str]]
Embeddings = collections.OrderedDict[str, list[float]]
EmbArr = numpy.typing.NDArray[numpy.float32]

class Scoring:
    # pylint: disable=too-many-instance-attributes
    """Computes scoring for a list of proteins.

    Args:
        embeddings: A vector to represent proteins.
        corr_threshold: Correlation threshold.
        corr_threshold_repl_value: The replacement value used in correlation
                                   function for values between corr_threshold
                                   and 1.0 - overlap_score_eps. To disable it,
                                   set it to 0.0.
        overlap_score: Beta value as used in the article.
        overlap_score_eps: To define a neighborhood around a maximum of correlation
    """

    # pylint: disable-next=too-many-positional-arguments
    def __init__(self, embeddings: Embeddings,
                 corr_threshold: float=0.95,
                 corr_threshold_repl_value: float=1.0,
                 overlap_score: float=3.0,
                 overlap_score_eps: float=1e-4) -> None:
        # pylint: disable=too-many-arguments

        self._embeddings = embeddings
        self._corr_threshold = corr_threshold
        self._corr_threshold_repl_value = corr_threshold_repl_value
        self._overlap_score = overlap_score
        self._overlap_score_eps = overlap_score_eps
        self._pathways : PathwaysDict | None = None
        self._pw_scores: dict[str, float] | None = None
        self._pw_emb: dict[str, EmbArr] | None = None

        # Check values
        if self._overlap_score < 1.0:
            raise ValueError("Overlap score must be greater or equal to 1.0.")

        self._check_embeddings()

    def _check_embeddings(self) -> None:

        # Check that the lists of floats have the same length
        ref_len = None
        for k, v in self._embeddings.items():
            if ref_len is None:
                ref_len = len(v)
            elif len(v) != ref_len:
                raise ValueError(f"Inside embeddings, item {k} has a different"
                                 + f"number of elements ({len(v)}) than other"
                                 + f"items ({ref_len}).")

    @property
    def corr_threshold(self) -> float:
        """Correlation threshold."""
        return self._corr_threshold

    @property
    def overlap_score(self) -> float:
        """Overlape score."""
        return self._overlap_score

    def comp_corr(self, a: EmbArr, b: EmbArr) -> float:
        """Computes correlation between two embeddings a and b.
        
        Args:
            a: First pathway embedding.
            b: Second pathway embedding.
            
        Returns:
            The computed correlation.
        """

        # Compute correlation (matrix nxn, n being the number of elements in a
        # or b.
        corr_dist = 1 - (scipy.spatial.distance.cdist(a, b, 'correlation'))

        # Remove interactions between genes inside the overlap region
        for idx, x in numpy.ndenumerate(corr_dist):
            if abs(x-1)<self._overlap_score_eps:
                corr_dist[idx[0], :] = 0
                corr_dist[:, idx[1]] = 0
                corr_dist[idx[0], idx[1]] = 1

        # Consider only max values (=> reduce dimension)
        corr_dist = corr_dist.max(axis=1)

        # Set small values to 0.0
        corr_dist[corr_dist < self._corr_threshold] = 0.0

        # Apply overlap score
        corr_dist[corr_dist >= (1.0 - self._overlap_score_eps)] = \
                self.overlap_score

        # Set in-between values to fixed value
        if self._corr_threshold_repl_value > 0.0:
            corr_dist[(corr_dist >= self._corr_threshold)
                      & (corr_dist < (1.0 - self._overlap_score_eps))] = \
                              self._corr_threshold_repl_value

        return float(numpy.mean(corr_dist))

    def get_items_embeddings(self, items: set[str]) -> EmbArr:
        """Select a set of items inside the provided embeddings.
        
        Args:
            items: The set of items for which to retrieve embeddings.
            
        Returns:
            The selected embeddings.
        """
        logger.debug("Build items embeddings.")
        selected_items = filter(lambda i: i in items, self._embeddings.keys())
        embeddings = numpy.array([self._embeddings[i] for i in selected_items],
                              dtype=numpy.float32)
        return embeddings

    def set_pathways(self, pathways :PathwaysDict) -> None:
        """Declares the pathways.
        
        Sets the pathways to use for computation.
        Side effect: previous computed pathway embeddings and pathway scores are
        reset.
        
        Args:
            pathways: dictionary with pathway IDs as keys and list of items as
            values.
        """
        self._pathways = pathways
        self._pw_emb = None
        self._pw_scores = None

    def get_pathway_keys(self) -> list[str]:
        """Gets the stored pathway IDs.
        
        Returns:
            The list of pathway IDs.
        """
        if self._pathways is None:
            raise RuntimeError("No pathways have been set.")
        return list(self._pathways.keys())

    def get_pathway_embeddings(self) -> dict[str, EmbArr]:
        """Gets the computed pathway embeddings.
        
        Computes the pathway embeddings, base on the provided embeddings and the
        list of pathways.
        
        Returns:
            A dictionary with pathway IDs as keys and embedding arrays as
            values.
        """

        if self._pw_emb is None:
            logger.debug("Build pathway embeddings.")
            if self._pathways is None:
                raise RuntimeError("Please, first set the pathways with the" +
                                   " set_pathways() method.")
            self._pw_emb = {}
            for key, items in self._pathways.items():
                self._pw_emb[key] = self.get_items_embeddings(set(items))

        return self._pw_emb

    def get_pathway_embedding(self, key: str) -> EmbArr:
        """Gets the embedding array of one particular pathway.
        
        Args:
            key: The ID of the targeted pathway.
            
        Returns:
            The embedding array of the targeted pathway.
        """

        return self.get_pathway_embeddings()[key]

    def compute_pathway_score(self, key: str, embeddings: EmbArr) -> float:
        """Computes the score of one pathway.
        
        Args:
            key: The ID of the targeted pathway.
            embeddings: The embeddings array with which to compute correlation.
        
        Returns:
            The correlation between the pathway's embeddings and the provided
            embeddings.
        """
        return self.comp_corr(embeddings, self.get_pathway_embedding(key))

    def compute_pathway_scores(self, embeddings: EmbArr) -> None:
        """Computes the scores of all pathways.
        
        Computes for each pathway, the correlation between its embeddings and
        the provided embeddings.

        Args:
            embeddings: The embeddings array with which to compute correlation.
            
        Returns:
            Nothing.
        """
        self._pw_scores = {}
        logger.debug("Compute pathways' scores.")
        for key, emb in tqdm.tqdm(self.get_pathway_embeddings().items(),
                                  desc = "Pathways scoring"):
            self._pw_scores[key] = self.comp_corr(embeddings, emb)

    def get_computed_pathway_scores(self) -> typing.Optional[dict[str, float]]:
        """Returns the pathway scores computed by compute_pathway_scores().
        
        Returns:
            The scores.
        """
        return self._pw_scores
